from __future__ import annotations

from typing import Any


def get_argument(
    args: tuple[Any, ...], kwargs: dict[str, Any], arg_name: str, arg_idx: int
) -> Any:
    """
    Get the argument both when it's passed as a positional argument or as a
    keyword argument
    """
    if kwargs.get(arg_name) is not None:
        return kwargs.get(arg_name)

    if len(args) > arg_idx:
        return args[arg_idx]

    return None



# pylint: disable=duplicate-code
from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Tuple, cast
from uuid import UUID

import llama_index
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    CompletionResponse,
    MessageRole,
)
from llama_index.core.base.response.schema import Response, StreamingResponse
from llama_index.core.callbacks import CBEventType, EventPayload
from llama_index.core.callbacks.base_handler import BaseCallbackHandler
from llama_index.core.chat_engine.types import (
    AgentChatResponse,
    StreamingAgentChatResponse,
)

from nebuly.entities import HistoryEntry, InteractionWatch, SpanWatch
from nebuly.events import Event, EventData, EventsStorage
from nebuly.requests import post_message

orig_download_loader = llama_index.core.readers.download_loader


def download_loader_patched(*args: Any, **kwargs: Any) -> Any:
    res = orig_download_loader(*args, **kwargs)
    orig_load = res.load_data

    def load_data(self: Any, *args: Any, **kwargs: Any) -> Any:
        res = orig_load(self, *args, **kwargs)
        for r in res:
            r.metadata["nebuly_rag_source"] = self.__class__.__name__
        return res

    res.load_data = load_data
    return res


# Patch download_loader to add the source of the RAG node in the metadata
llama_index.core.readers.download_loader = download_loader_patched


def _get_spans(
    events: dict[UUID, Event], trace_map: dict[str, list[str]] | None = None
) -> list[SpanWatch]:
    if trace_map is None:
        event_ids = list(events.keys())
    else:
        values = trace_map.values()
        event_ids = [UUID(item) for sublist in values for item in sublist]
    return [
        event.as_span_watch()
        for event in events.values()
        if event.event_id in event_ids
    ]


@dataclass
class LLamaIndexEvent(Event):
    @property
    def input(self) -> str:
        if self.data.kwargs is None:
            return ""

        input_payload = self.data.kwargs["input_payload"]
        if CBEventType(self.data.type) is CBEventType.QUERY:
            return str(input_payload[EventPayload.QUERY_STR])
        if CBEventType(self.data.type) in [CBEventType.LLM, CBEventType.AGENT_STEP]:
            if EventPayload.MESSAGES in input_payload:
                last_user_input = input_payload[EventPayload.MESSAGES][-1]
                if isinstance(last_user_input, ChatMessage):
                    return str(last_user_input.content)
                return str(last_user_input)
            return str(input_payload[EventPayload.PROMPT])

        return ""

    @property
    def output(self) -> str:
        if self.data.kwargs is None:
            return ""

        if self.data.output:
            return str(self.data.output)
        output = self.data.kwargs["output_payload"][EventPayload.RESPONSE]
        if hasattr(output, "response") and output.response:
            return str(output.response)
        if isinstance(output, StreamingResponse):
            return str(output.get_response().response)

        return ""

    @property
    def history(self) -> list[HistoryEntry]:
        if self.data.kwargs is None:
            return []

        if CBEventType(self.data.type) is CBEventType.LLM:
            input_payload = self.data.kwargs.get("input_payload")
            if input_payload is not None:
                messages: list[ChatMessage] = input_payload.get(EventPayload.MESSAGES)
                if messages is not None:
                    messages = [
                        m
                        for m in messages
                        if m.role in [MessageRole.USER, MessageRole.ASSISTANT]
                    ]
                    if len(messages) % 2 != 0:
                        return [
                            HistoryEntry(
                                user=str(messages[i].content),
                                assistant=str(messages[i + 1].content),
                            )
                            for i in range(len(messages[:-1]) - 1)
                        ]
        return []

    def _get_rag_source(self) -> str | None:
        if self.data.kwargs is None:
            return None

        if CBEventType(self.data.type) in [CBEventType.QUERY, CBEventType.RETRIEVE]:
            if CBEventType(self.data.type) is CBEventType.QUERY:
                response = self.data.kwargs["output_payload"][EventPayload.RESPONSE]
                source_nodes = getattr(response, "source_nodes", [])
            else:
                source_nodes = self.data.kwargs["output_payload"][EventPayload.NODES]
            if len(source_nodes) > 0:
                file_name = source_nodes[0].metadata.get("file_name", None)
                if file_name is not None:
                    return str(file_name)
                rag_source: str | None = source_nodes[0].metadata.get(
                    "nebuly_rag_source", source_nodes[0].node_id
                )
                return rag_source
        if CBEventType(self.data.type) is CBEventType.FUNCTION_CALL:
            function_name: str | None = self.data.kwargs["input_payload"][
                EventPayload.TOOL
            ].name
            return function_name

        return None

    def as_span_watch(self) -> SpanWatch:
        if self.end_time is None:
            self.set_end_time()
        return SpanWatch(
            span_id=self.event_id,
            module="llama_index",
            version="unknown",
            function=self.data.type,
            called_start=self.start_time,
            called_end=cast(datetime, self.end_time),
            called_with_args=cast(Tuple[Any], self.data.args),
            called_with_kwargs=cast(Dict[str, Any], self.data.kwargs),
            returned=self.data.output,
            generator=False,
            generator_first_element_timestamp=None,
            provider_extras={
                "event_type": self.data.type,
            },
            rag_source=self._get_rag_source(),
        )


class LlamaIndexTrackingHandler(
    BaseCallbackHandler
):  # pylint: disable=too-many-instance-attributes
    def __init__(  # pylint: disable=too-many-arguments
        self,
        api_key: str,
        user_id: str,
        user_group_profile: str | None = None,
        nebuly_tags: dict[str, str] | None = None,
        feature_flags: list[str] | None = None,
    ) -> None:
        super().__init__(event_starts_to_ignore=[], event_ends_to_ignore=[])
        self.api_key = api_key
        self.nebuly_user = user_id
        self.nebuly_user_group = user_group_profile
        self.tags = nebuly_tags
        self._events_storage = EventsStorage()

        # Attributes needed for handling streaming responses
        self._waiting_stream_response = False
        self._trace_map: dict[str, list[str]] | None = None
        self._trace_id: str | None = None
        self.nebuly_feature_flags = feature_flags

    @staticmethod
    def _trace_map_to_hierarchy(
        trace_map: dict[str, list[str]] | None, run_id: uuid.UUID
    ) -> dict[UUID, UUID | None]:
        hierarchy = {}
        if trace_map is not None:
            for parent_id, child_ids in trace_map.items():
                if parent_id == "root":
                    p_id = None
                else:
                    p_id = UUID(parent_id)
                for child_id in child_ids:
                    hierarchy[UUID(child_id)] = p_id
        else:
            hierarchy[run_id] = None
        return hierarchy

    def _get_event_history(
        self, trace_id: str | None, event: Event, spans: list[SpanWatch]
    ) -> list[HistoryEntry]:
        if trace_id != "chat" or CBEventType(event.data.type) is CBEventType.LLM:
            return event.history
        # If the event is a chat event, and the event is not a LLM event, we need to
        # find the first event containing the LLM response to get the history
        for span in spans:
            if span.function == "llm":
                llm_event = self._events_storage.events[span.span_id]
                return llm_event.history
        return []

    def _send_interaction(
        self,
        trace_id: str | None,
        root_id: uuid.UUID,
        trace_map: dict[str, list[str]] | None,
        final_stream_event_run_id: uuid.UUID | None = None,
    ) -> None:
        if (
            len(self._events_storage.events) < 1
            or root_id not in self._events_storage.events
        ):
            raise ValueError(f"Event {root_id} not found in events storage.")

        event = self._events_storage.events[root_id]
        if final_stream_event_run_id is not None:
            response = self._events_storage.events[  # type: ignore
                final_stream_event_run_id
            ].data.kwargs["output_payload"][EventPayload.RESPONSE]
            output = response.message.content
        else:
            output = event.output
        spans = _get_spans(events=self._events_storage.events, trace_map=trace_map)
        interaction = InteractionWatch(
            end_user=self.nebuly_user,
            end_user_group_profile=self.nebuly_user_group,
            input=event.input,
            output=output,
            time_start=event.start_time,
            time_end=cast(datetime, event.end_time),
            history=self._get_event_history(trace_id, event, spans),
            spans=spans,
            hierarchy=self._trace_map_to_hierarchy(trace_map=trace_map, run_id=root_id),
            tags=self.tags,
            feature_flags=self.nebuly_feature_flags,
        )
        post_message(interaction, self.api_key)

    def start_trace(self, trace_id: str | None = None) -> None:
        return

    def end_trace(
        self,
        trace_id: str | None = None,
        trace_map: dict[str, list[str]] | None = None,
        final_stream_event_run_id: uuid.UUID | None = None,
    ) -> None:
        if (
            trace_id in ["query", "chat"]
            and trace_map is not None
            and "root" in trace_map
        ):
            valid_root_ids = [
                event_id
                for event_id in trace_map["root"]
                if self._events_storage.events[UUID(event_id)].data.type
                in ["query", "agent_step", "llm"]
            ]
            if len(valid_root_ids) == 0:
                return
            root_id = UUID(valid_root_ids[0])
            if final_stream_event_run_id is not None:
                event = self._events_storage.events[final_stream_event_run_id]
            else:
                event = self._events_storage.events[root_id]
            response = event.data.kwargs.get("output_payload", {}).get(  # type: ignore
                EventPayload.RESPONSE, None
            )
            if response is not None:
                if isinstance(
                    response, (StreamingResponse, StreamingAgentChatResponse)
                ):
                    # When streaming responses are used, the end_trace is called before
                    # the end_event, so we need store some variables to send the
                    # interaction when the end_event will be received later
                    self._waiting_stream_response = True
                    self._trace_map = trace_map
                    self._trace_id = trace_id
                else:
                    self._send_interaction(
                        trace_id,
                        root_id,
                        trace_map=trace_map,
                        final_stream_event_run_id=final_stream_event_run_id,
                    )
                    self._events_storage.delete_events(root_id)

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: dict[str, Any] | None = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> str:
        if payload is not None:
            self._events_storage.add_event(
                constructor=LLamaIndexEvent,
                event_id=UUID(event_id),
                parent_id=UUID(parent_id)
                if parent_id and parent_id != "root"
                else None,
                data=EventData(
                    type=event_type.value,
                    args=tuple(),
                    kwargs={"input_payload": payload, **kwargs},
                ),
                module="llama_index",
            )
        return event_id

    def on_event_end(  # pylint: disable=too-many-branches
        self,
        event_type: CBEventType,
        payload: dict[str, Any] | None = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        run_id = UUID(event_id)
        if payload is not None:
            if event_type in [
                CBEventType.QUERY,
                CBEventType.LLM,
                CBEventType.AGENT_STEP,
            ]:
                if EventPayload.RESPONSE in payload:
                    response = payload[EventPayload.RESPONSE]
                elif EventPayload.COMPLETION in payload:
                    response = payload[EventPayload.COMPLETION]
                else:
                    raise ValueError("Unknown response type.")
                if isinstance(response, (Response, AgentChatResponse)):
                    output = response.response
                elif isinstance(response, ChatResponse):
                    output = response.message.content
                elif isinstance(response, CompletionResponse):
                    output = response.text
                elif isinstance(
                    response, (StreamingResponse, StreamingAgentChatResponse)
                ):
                    output = None
                else:
                    raise ValueError(f"Unknown response type: {type(response)}")
            else:
                output = None

            self._events_storage.events[run_id].data.add_end_event_data(
                kwargs={"output_payload": payload, **kwargs}, output=output
            )
            self._events_storage.events[run_id].set_end_time()

            if event_type is CBEventType.LLM and len(self._events_storage.events) == 1:
                # Track single LLM calls outside of a trace
                self._send_interaction(trace_id=None, root_id=run_id, trace_map=None)
                self._events_storage.delete_events(run_id)

            if self._waiting_stream_response:
                # When streaming responses are used, the end_trace is called before
                # the end_event, so we need to wait for the end_event to send the
                # interaction
                self._waiting_stream_response = False
                self.end_trace(
                    self._trace_id,
                    trace_map=self._trace_map,
                    final_stream_event_run_id=run_id,
                )
                self._trace_map = None
                self._trace_id = None



from __future__ import annotations

from functools import partial
from queue import Queue

from nebuly.api_key import get_api_key
from nebuly.config import PACKAGES
from nebuly.consumers import ConsumerWorker
from nebuly.entities import InteractionWatch, Observer
from nebuly.exceptions import NebulyAlreadyInitializedError
from nebuly.monkey_patching import (
    check_no_packages_already_imported,
    import_and_patch_packages,
)
from nebuly.observers import NebulyObserver
from nebuly.requests import post_message

_initialized = False


def init(
    *,
    api_key: str | None = None,
    anonymize: bool = True,
) -> None:
    if not api_key:
        api_key = get_api_key()

    _check_nebuly_is_not_initialized()
    check_no_packages_already_imported(PACKAGES)
    observer = _create_observer_and_start_publisher(
        api_key=api_key, anonymize=anonymize
    )
    import_and_patch_packages(PACKAGES, observer)


def _check_nebuly_is_not_initialized() -> None:
    global _initialized  # pylint: disable=global-statement
    if _initialized:
        raise NebulyAlreadyInitializedError("Nebuly already initialized")
    _initialized = True


def _create_observer_and_start_publisher(
    *, api_key: str, anonymize: bool = True
) -> Observer:
    queue: Queue[InteractionWatch] = Queue()

    ConsumerWorker(queue, partial(post_message, api_key=api_key, anonymize=anonymize))
    observer = NebulyObserver(
        api_key=api_key,
        publish=queue.put,
    )
    return observer.on_event_received



# pylint: disable=duplicate-code
import asyncio
import json
import os
from abc import ABC
from datetime import datetime, timezone
from typing import Any, Optional

from nebuly.api_key import get_api_key
from nebuly.requests import post_json_data
from nebuly.tracking.types import (
    FeedbackAction,
    FeedbackActionMetadata,
    FeedbackActionName,
)


class TrackingBase(ABC):
    """
    Base class for ABTesting and AsyncABTesting
    """

    def __init__(self, api_key: Optional[str] = None, anonymize: bool = True) -> None:
        if api_key is None:
            api_key = get_api_key()
        self._api_key = api_key
        url = (
            os.environ.get("NEBULY_API_URL")
            or "https://backend.nebuly.com/event-ingestion/api/v1/events"
        )
        self._URL = f"{url}/feedback"
        self._anonymize = anonymize

    def get_request_payload(  # pylint: disable=too-many-arguments
        self,
        user_id: str,
        action: str,
        text: Optional[str] = None,
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
        output: Optional[str] = None,
    ) -> str:
        feedback_action = FeedbackAction(
            slug=FeedbackActionName(action),
            extras={"text": text} if text else None,
        )
        action_metadata = FeedbackActionMetadata(
            input=input,
            output=output,
            end_user=user_id,
            timestamp=datetime.now(timezone.utc),
            anonymize=self._anonymize,
        )
        request_data = {
            "action": feedback_action.to_dict(),
            "metadata": action_metadata.to_dict(),
        }
        payload = json.dumps(request_data)
        return payload


class TrackingSDK(TrackingBase):
    """
    Synchronous Feedback Action tracking class
    """

    def send_feedback_action(  # pylint: disable=too-many-arguments
        self,
        user_id: str,
        action: str,
        text: Optional[str] = None,
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
        output: Optional[str] = None,
    ) -> Any:
        """
        Send the feedback action to Nebuly's platform
        """
        payload = self.get_request_payload(user_id, action, text, input, output)
        return post_json_data(self._URL, payload, self._api_key)


class AsyncTrackingSDK(TrackingBase):
    """
    Asynchronous Feedback Action tracking class
    """

    async def send_feedback_action(  # pylint: disable=too-many-arguments
        self,
        user_id: str,
        action: str,
        text: Optional[str] = None,
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
        output: Optional[str] = None,
    ) -> Any:
        """
        Send the feedback action to Nebuly's platform
        """
        payload = self.get_request_payload(user_id, action, text, input, output)
        loop = asyncio.get_running_loop()
        raw_response = await loop.run_in_executor(
            None,
            post_json_data,
            self._URL,
            payload,
            self._api_key,
        )
        return raw_response



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Callable
from uuid import UUID

from nebuly.entities import EventHierarchy, HistoryEntry, SpanWatch


@dataclass
class EventData:
    type: Any
    args: tuple[Any, ...] | None = None
    kwargs: dict[str, Any] | None = None
    output: Any | None = None

    def add_end_event_data(
        self,
        output: Any,
        args: tuple[Any, ...] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None:
        if self.args is None:
            self.args = tuple()
        if self.kwargs is None:
            self.kwargs = {}
        self.args += args if args is not None else tuple()
        self.kwargs = dict(self.kwargs, **kwargs) if kwargs is not None else self.kwargs
        self.output = output


@dataclass
class Event(abc.ABC):
    event_id: UUID
    hierarchy: EventHierarchy | None
    data: EventData
    module: str
    start_time: datetime
    end_time: datetime | None = None

    @property
    @abc.abstractmethod
    def input(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def output(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def history(self) -> list[HistoryEntry]:
        raise NotImplementedError

    @abc.abstractmethod
    def as_span_watch(self) -> SpanWatch:
        raise NotImplementedError

    def set_end_time(self) -> None:
        self.end_time = datetime.now(timezone.utc)


@dataclass
class EventsStorage:
    events: dict[UUID, Event] = field(default_factory=dict)

    def get_root_id(self, event_id: UUID) -> UUID | None:
        if event_id not in self.events:
            return None
        event = self.events[event_id]
        if event.hierarchy is None:
            return event_id
        return self.get_root_id(event.hierarchy.root_run_id)

    def add_event(  # pylint: disable=too-many-arguments
        self,
        constructor: Callable[..., Event],
        event_id: UUID,
        parent_id: UUID | None,
        data: EventData,
        module: str,
    ) -> None:
        if event_id in self.events:
            raise ValueError(f"Event {event_id} already exists in events storage.")

        # Handle new event
        if parent_id is None:
            hierarchy = None
        else:
            root_id = self.get_root_id(parent_id)
            if root_id is None:
                hierarchy = None
            else:
                hierarchy = EventHierarchy(parent_run_id=parent_id, root_run_id=root_id)
        self.events[event_id] = constructor(
            event_id=event_id,
            hierarchy=hierarchy,
            data=data,
            module=module,
            start_time=datetime.now(timezone.utc),
        )

    def delete_events(self, root_id: UUID) -> None:
        if root_id not in self.events:
            raise ValueError(f"Event {root_id} not found in events hierarchy storage.")

        keys_to_delete = [root_id]

        for event_id, event_detail in self.events.items():
            if (
                event_detail.hierarchy is not None
                and event_detail.hierarchy.root_run_id == root_id
            ):
                keys_to_delete.append(event_id)

        for key in keys_to_delete:
            self.events.pop(key)



import importlib.metadata

from packaging import version


def is_module_version_less_than(module_name: str, version_spec: str) -> bool:
    """
    Check if the installed version of a module is less than the specified version.

    Parameters:
    module_name (str): The name of the module to check.
    version_spec (str): The version to compare against in the format
    'major.minor.patch'.

    Returns:
    bool: True if the installed version is less than the specified version,
    False otherwise.
    """
    try:
        installed_version = importlib.metadata.version(module_name)
        return version.parse(installed_version) < version.parse(version_spec)
    except importlib.metadata.PackageNotFoundError:
        print(f"Module '{module_name}' is not installed.")
        return False



import os

from nebuly.exceptions import APIKeyNotProvidedError


def get_api_key() -> str:
    api_key = os.environ.get("NEBULY_API_KEY")
    if not api_key:
        raise APIKeyNotProvidedError(
            "API key not provided and not found in environment"
        )
    return api_key



from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable


@dataclass(frozen=True)
class EventHierarchy:
    parent_run_id: uuid.UUID
    root_run_id: uuid.UUID


@dataclass(frozen=True)
class SupportedVersion:
    min_version: str
    max_version: str | None = None


@dataclass(frozen=True)
class Package:
    """
    Package represents a package to be patched.
    """

    name: str
    versions: SupportedVersion
    to_patch: tuple[str, ...]


@dataclass(frozen=True)
class HistoryEntry:
    """
    HistoryEntry represents a user/assistant interaction in a chat history.
    """

    user: str
    assistant: str

    def to_dict(self) -> list[str]:
        """
        to_dict returns a dictionary representation of the HistoryEntry instance.
        """
        return [self.user, self.assistant]


@dataclass(frozen=True)
class ModelInput:
    prompt: str
    history: list[HistoryEntry] = field(default_factory=list)


@dataclass
class SpanWatch:  # pylint: disable=too-many-instance-attributes
    """
    SpanWatch represents a call to a function that was patched.
    """

    module: str
    version: str
    function: str
    called_start: datetime
    called_end: datetime
    called_with_args: tuple[Any, ...]
    called_with_kwargs: dict[str, Any]
    returned: Any
    generator: bool
    generator_first_element_timestamp: datetime | None
    provider_extras: dict[str, Any] | None = None
    rag_source: str | None = None
    media: list[str] | None = None
    span_id: uuid.UUID = field(default_factory=uuid.uuid4)

    @staticmethod
    def clean_kwargs(kwargs: dict[str, Any]) -> dict[str, Any]:
        # Find invocation_params if present and remove nebuly_interaction
        if "invocation_params" in kwargs:
            if "nebuly_interaction" in kwargs["invocation_params"]:
                kwargs["invocation_params"].pop("nebuly_interaction")

        return kwargs

    def to_dict(self) -> dict[str, Any]:
        """
        to_dict returns a dictionary representation of the SpanWatch instance.
        """
        return {
            "module": self.module,
            "version": self.version,
            "function": self.function,
            "called_start": self.called_start.isoformat(),
            "called_end": self.called_end.isoformat(),
            "called_with_args": self.called_with_args,
            "called_with_kwargs": self.clean_kwargs(self.called_with_kwargs),
            "returned": self.returned,
            "generator": self.generator,
            "generator_first_element_timestamp": (
                self.generator_first_element_timestamp.isoformat()
                if self.generator_first_element_timestamp
                else None
            ),
            "provider_extras": {
                k: v
                for k, v in self.provider_extras.items()
                if k != "nebuly_interaction"
            }
            if self.provider_extras is not None
            else None,
            "rag_source": self.rag_source,
            "span_id": str(self.span_id),
            "media": self.media,
        }


@dataclass
class InteractionWatch:  # pylint: disable=too-many-instance-attributes
    input: str
    output: str
    time_end: datetime
    time_start: datetime
    spans: list[SpanWatch]
    history: list[HistoryEntry]
    hierarchy: dict[uuid.UUID, uuid.UUID | None]
    end_user: str
    end_user_group_profile: str | None
    tags: dict[str, str] | None = None
    feature_flags: list[str] | None = None
    api_key: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """
        to_dict returns a dictionary representation of the InteractionWatch instance.
        """
        return {
            "input": self.input,
            "output": self.output,
            "time_end": self.time_end.isoformat(),
            "time_start": self.time_start.isoformat(),
            "spans": [span.to_dict() for span in self.spans],
            "history": self.history,
            "hierarchy": {
                str(k): str(v) if v is not None else None
                for k, v in self.hierarchy.items()
            },
            "end_user": self.end_user,
            "end_user_group_profile": self.end_user_group_profile,
            "tags": self.tags,
            "feature_flags": self.feature_flags,
        }


Observer = Callable[[InteractionWatch], None]

Publisher = Callable[[InteractionWatch], None]



from __future__ import annotations

from typing import Any, Callable

from anthropic import Anthropic, AsyncAnthropic, AsyncStream, Stream
from anthropic.types import Completion

from nebuly.entities import ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.providers.common import extract_anthropic_input_and_history


def is_anthropic_generator(function: Callable[[Any], Any]) -> bool:
    return isinstance(function, (Stream, AsyncStream))


class AnthropicPicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        names = [
            "auth_token",
            "base_url",
            "api_key",
            "timeout",
            "max_retries",
            "default_headers",
        ]
        return {
            key: {
                "key_names": names,
                "attribute_names": names,
            }
            for key in [Anthropic, AsyncAnthropic]
        }


def handle_anthropic_unpickable_objects() -> None:
    pickler_handler = AnthropicPicklerHandler()
    for obj in [Anthropic, AsyncAnthropic]:
        pickler_handler.handle_unpickable_object(
            obj=obj,
        )


class AnthropicDataExtractor(ProviderDataExtractor):
    def __init__(
        self,
        function_name: str,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
    ):
        self.function_name = function_name
        self.original_args = original_args
        self.original_kwargs = original_kwargs

    def extract_input_and_history(self, outputs: Any) -> ModelInput:
        if self.function_name in [
            "resources.Completions.create",
            "resources.AsyncCompletions.create",
        ]:
            last_user_input, history = extract_anthropic_input_and_history(
                self.original_kwargs["prompt"]
            )
            return ModelInput(prompt=last_user_input, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: Completion | list[Completion],
    ) -> str:
        if stream and isinstance(outputs, list):
            return self._extract_output_generator(outputs)
        if isinstance(outputs, Completion) and self.function_name in [
            "resources.Completions.create",
            "resources.AsyncCompletions.create",
        ]:
            return outputs.completion.strip()

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )

    def _extract_output_generator(self, outputs: list[Completion]) -> str:
        if all(
            isinstance(output, Completion) for output in outputs
        ) and self.function_name in [
            "resources.Completions.create",
            "resources.AsyncCompletions.create",
        ]:
            return "".join([output.completion for output in outputs]).strip()

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Literal, Sequence, TypedDict, Union

from typing_extensions import TypeAlias


class VariantKind(str, Enum):
    PROMPT = "prompt"
    MODEL = "model"
    RAG_SOURCE = "rag_source"


class ConfigParamType(str, Enum):
    STRING = "string"
    NUMBER = "number"
    BOOLEAN = "boolean"
    ARRAY = "array"


@dataclass
class ConfigParam:
    name: str
    value: Any
    param_type: ConfigParamType

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ConfigParam":
        if "name" not in data or "value" not in data or "param_type" not in data:
            raise ValueError("Could not parse ConfigParam")
        return cls(
            name=data["name"],
            value=data["value"],
            param_type=ConfigParamType(data["param_type"]),
        )


@dataclass
class PromptVariant:
    prompt: str
    config_params: List[ConfigParam]
    kind: Literal[VariantKind.PROMPT] = VariantKind.PROMPT

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "PromptVariant":
        if "prompt" not in data or "config_params" not in data:
            raise ValueError("Could not parse PromptVariant")
        return cls(
            prompt=data["prompt"],
            config_params=[ConfigParam(**param) for param in data["config_params"]],
        )


@dataclass
class ModelVariant:
    model_id: str
    config_params: List[ConfigParam]
    kind: Literal[VariantKind.MODEL] = VariantKind.MODEL

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ModelVariant":
        if "model_id" not in data or "config_params" not in data:
            raise ValueError("Could not parse ModelVariant")
        return cls(
            model_id=data["model_id"],
            config_params=[ConfigParam(**param) for param in data["config_params"]],
        )


@dataclass
class RagSourceVariant:
    rag_source: str
    config_params: List[ConfigParam]
    kind: Literal[VariantKind.RAG_SOURCE] = VariantKind.RAG_SOURCE

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "RagSourceVariant":
        if "rag_source" not in data or "config_params" not in data:
            raise ValueError("Could not parse RagSourceVariant")
        return cls(
            rag_source=data["rag_source"],
            config_params=[ConfigParam(**param) for param in data["config_params"]],
        )


Variant: TypeAlias = Union[PromptVariant, ModelVariant, RagSourceVariant]


def variant_from_dict(data: Dict[str, Any]) -> Variant:
    if "kind" not in data:
        raise ValueError("Could not parse variant")
    if data["kind"] == VariantKind.PROMPT:
        return PromptVariant.from_dict(data)
    if data["kind"] == VariantKind.MODEL:
        return ModelVariant.from_dict(data)
    if data["kind"] == VariantKind.RAG_SOURCE:
        return RagSourceVariant.from_dict(data)
    raise ValueError("Could not parse variant")


@dataclass
class LookupVariant:
    feature_flag_name: str
    variant: Variant

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LookupVariant":
        if "feature_flag_name" not in data or "variant" not in data:
            raise ValueError("Could not parse LookupVariant")
        return cls(
            feature_flag_name=data["feature_flag_name"],
            variant=variant_from_dict(data["variant"]),
        )


@dataclass
class Response:
    variants: List[LookupVariant]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Response":
        if "variants" not in data:
            raise ValueError("Could not parse response")
        return cls(
            variants=[LookupVariant.from_dict(variant) for variant in data["variants"]]
        )


class Request(TypedDict):
    user: str
    feature_flags: Sequence[str]



from __future__ import annotations

import inspect
from contextlib import contextmanager
from datetime import datetime, timezone
from typing import Dict, Generator, List, cast
from uuid import UUID

from nebuly.entities import HistoryEntry, InteractionWatch, Observer, SpanWatch
from nebuly.exceptions import (
    AlreadyInInteractionContext,
    InteractionContextInitiationError,
    InteractionMustBeLocalVariable,
    MissingRequiredNebulyFieldError,
    NotInInteractionContext,
)


class InteractionContext:  # pylint: disable=too-many-instance-attributes
    def __init__(  # pylint: disable=too-many-arguments
        self,
        user: str | None = None,
        user_group_profile: str | None = None,
        initial_input: str | None = None,
        final_output: str | None = None,
        history: list[HistoryEntry] | None = None,
        hierarchy: dict[UUID, UUID | None] | None = None,
        spans: list[SpanWatch] | None = None,
        do_not_call_directly: bool = False,
        tags: dict[str, str] | None = None,
        feature_flags: list[str] | None = None,
        nebuly_api_key: str | None = None,
    ) -> None:
        if not do_not_call_directly:
            raise InteractionContextInitiationError(
                "Interaction cannot be directly instantiate, use the"
                " 'new_interaction' contextmanager"
            )
        self._observer: Observer | None = None
        self._finished: bool = False

        self.user = user
        self.user_group_profile = user_group_profile
        self.input = initial_input
        self.output = final_output
        self.spans = [] if spans is None else spans
        self.history = history
        self.hierarchy = hierarchy
        self.time_start = datetime.now(timezone.utc)
        self.tags = tags
        self.feature_flags = feature_flags
        self.nebuly_api_key = nebuly_api_key

    def set_input(self, value: str) -> None:
        self.input = value

    def set_history(self, value: list[dict[str, str]] | list[HistoryEntry]) -> None:
        if len(value) == 0:
            self.history = []
        elif isinstance(value[0], dict):
            value = cast(List[Dict[str, str]], value)
            history = [
                message
                for message in value
                if message["role"].lower() in ["user", "assistant"]
            ]
            if len(history) % 2 != 0:
                raise ValueError(
                    "Odd number of chat history elements, please provide "
                    "a valid history."
                )
            self.history = [
                HistoryEntry(
                    user=history[i]["content"], assistant=history[i + 1]["content"]
                )
                for i in range(len(history) - 1)
            ]
        else:
            value = cast(List[HistoryEntry], value)
            self.history = value

    def set_output(self, value: str) -> None:
        self.output = value

    def _set_observer(self, observer: Observer) -> None:
        self._observer = observer

    def _add_span(self, value: SpanWatch) -> None:
        self.spans.append(value)

    def _finish(self) -> None:
        self._finished = True
        self.hierarchy = {}
        for span in self.spans:
            if span.provider_extras is None:
                continue
            parent_id = span.provider_extras.get("parent_run_id")
            if parent_id is not None:
                self.hierarchy[span.span_id] = UUID(parent_id)
        if self._observer is not None:
            self._observer(self._as_interaction_watch())

    def _set_user(self, value: str) -> None:
        self.user = value

    def _set_user_group_profile(self, value: str | None) -> None:
        self.user_group_profile = value

    def _add_tags(self, tags: dict[str, str]) -> None:
        if self.tags is None:
            self.tags = {}
        self.tags = {**self.tags, **tags}

    def _add_feature_flags(self, flags: list[str]) -> None:
        if self.feature_flags is None:
            self.feature_flags = []
        self.feature_flags.extend(flags)

    def _set_api_key(self, api_key: str) -> None:
        self.nebuly_api_key = api_key

    def _validate_interaction(self) -> None:
        if self.input is None:
            raise ValueError("Interaction has no input.")
        if self.output is None:
            raise ValueError("Interaction has no output.")
        if self.history is None:
            raise ValueError("Interaction has no history.")
        if self.hierarchy is None:
            raise ValueError("Interaction has no hierarchy.")
        if self.hierarchy:
            for value in self.hierarchy.values():
                if value is not None and value not in self.hierarchy:
                    raise ValueError(
                        f"Interaction hierarchy has a reference to a non-existing "
                        f"event: {value}"
                    )
        if self.user is None:
            raise MissingRequiredNebulyFieldError(
                "Missing required nebuly field: 'user_id'. Please add it when calling "
                "the original provider method."
            )

    def _as_interaction_watch(self) -> InteractionWatch:
        self._validate_interaction()
        return InteractionWatch(
            input=self.input,  # type: ignore
            output=self.output,  # type: ignore
            time_start=self.time_start,
            time_end=datetime.now(timezone.utc),
            spans=self.spans,
            history=self.history,  # type: ignore
            hierarchy=self.hierarchy,  # type: ignore
            end_user=self.user,  # type: ignore
            end_user_group_profile=self.user_group_profile,
            tags=self.tags,
            feature_flags=self.feature_flags,
            api_key=self.nebuly_api_key,
        )


def get_nearest_open_interaction() -> InteractionContext:
    frames = inspect.stack()
    for frame in frames[::-1]:
        for v in frame.frame.f_locals.values():
            if (
                isinstance(v, InteractionContext)
                and not v._finished  # pylint: disable=protected-access
            ):
                return v
    raise NotInInteractionContext()


@contextmanager
def new_interaction(
    user_id: str | None = None,
    user_group_profile: str | None = None,
    nebuly_tags: dict[str, str] | None = None,
    auto_publish: bool = True,
    feature_flags: list[str] | None = None,
) -> Generator[InteractionContext, None, None]:
    try:
        get_nearest_open_interaction()
        raise AlreadyInInteractionContext()
    except NotInInteractionContext:
        pass

    try:
        yield InteractionContext(
            user_id,
            user_group_profile,
            tags=nebuly_tags,
            feature_flags=feature_flags,
            do_not_call_directly=True,
        )
    finally:
        if auto_publish:
            try:
                interaction = get_nearest_open_interaction()
            except NotInInteractionContext:
                raise InteractionMustBeLocalVariable()  # pylint: disable=raise-missing-from  # noqa: E501
            interaction._finish()  # pylint: disable=protected-access



# pylint: disable=duplicate-code, import-error, no-name-in-module
# mypy: ignore-errors
from __future__ import annotations

import copyreg
import json
import logging
from typing import Any, Callable

from botocore.eventstream import EventStream
from botocore.response import StreamingBody

from nebuly.entities import ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.providers.common import extract_anthropic_input_and_history

logger = logging.getLogger(__name__)


def is_aws_bedrock_generator(obj: Any) -> bool:
    return isinstance(obj["body"], EventStream)


def is_model_supported(model_id: str) -> bool:
    if model_id.startswith("stability") or "-embed-" in model_id:
        return False
    return True


class AWSBedrockPicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        return {}

    def _pickle_object(self, obj: Any) -> Any:
        # Ignore the BotocoreRuntime object
        return lambda x: x, (None,)


def handle_aws_bedrock_unpickable_objects(obj: Any) -> None:
    pickler_handler = AWSBedrockPicklerHandler()
    pickler_handler.handle_unpickable_object(
        obj=obj.__class__,
    )

    def extract_aws_streaming_body(obj: StreamingBody) -> Any:  # type: ignore
        def new_read():
            return obj._raw_stream.data  # pylint: disable=protected-access

        data = obj._raw_stream.data  # pylint: disable=protected-access
        obj.read = new_read
        return lambda x: x, (data,)

    copyreg.pickle(StreamingBody, extract_aws_streaming_body)


class AWSBedrockDataExtractor(ProviderDataExtractor):
    def __init__(
        self,
        function_name: str,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
    ):
        self.function_name = function_name
        self.original_args = original_args
        self.original_kwargs = original_kwargs
        self.provider = original_args[2]["modelId"].split(".")[0]

    def extract_input_and_history(self, outputs: Any) -> ModelInput:
        if self.function_name == "client.BaseClient._make_api_call":
            if self.provider == "amazon":  # Amazon
                return ModelInput(
                    prompt=json.loads(self.original_args[2]["body"])["inputText"]
                )
            if self.provider == "anthropic":  # Anthropic
                last_user_input, history = extract_anthropic_input_and_history(
                    json.loads(self.original_args[2]["body"])["prompt"]
                )
                return ModelInput(prompt=last_user_input, history=history)
            # Cohere and AI21
            return ModelInput(
                prompt=json.loads(self.original_args[2]["body"])["prompt"]
            )

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self, stream: bool, outputs: dict[str, Any] | list[dict[str, Any]]
    ) -> str:
        if isinstance(outputs, list) and stream:
            return self._extract_output_generator(outputs)

        if (
            isinstance(outputs, dict)
            and self.function_name == "client.BaseClient._make_api_call"
        ):
            response_body = json.loads(outputs["body"].decode("utf-8"))
            if self.provider == "amazon":
                return response_body["results"][0]["outputText"]
            if self.provider == "cohere":
                return response_body["generations"][0]["text"]
            if self.provider == "anthropic":
                return response_body["completion"].strip()
            if self.provider == "ai21":
                return response_body["completions"][0]["data"]["text"]

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )

    def _extract_output_generator(self, outputs: list[dict[str, Any]]) -> str:
        result = ""
        for output in outputs:
            chunk = json.loads(output["chunk"]["bytes"].decode("utf-8"))
            # AI21 does not support streaming
            if self.provider == "amazon":
                result += chunk["outputText"]
            elif self.provider == "cohere":
                # Must check for "text" because it isn't present in the last chunks
                if "text" in chunk and chunk["text"] != "<EOS_TOKEN>":
                    result += chunk["text"]
            elif self.provider == "anthropic":
                result += chunk["completion"]
            else:
                raise ValueError(f"Provider {self.provider} not supported for stream")
        if self.provider == "anthropic":
            result = result.strip()
        return result



# pylint: disable=duplicate-code, import-error, no-name-in-module
# mypy: ignore-errors
from __future__ import annotations

import json
import logging
from copy import deepcopy
from typing import Any, Callable, List, Tuple, cast

import openai
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI, _ModuleClient
from openai._response import APIResponse
from openai.pagination import SyncCursorPage
from openai.types.chat.chat_completion import (  # type: ignore  # noqa: E501
    ChatCompletion,
    Choice,
)
from openai.types.chat.chat_completion_chunk import (  # type: ignore  # noqa: E501
    ChatCompletionChunk,
)
from openai.types.completion import Completion  # type: ignore  # noqa: E501
from openai.types.completion_choice import (  # type: ignore  # noqa: E501
    CompletionChoice,
)

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import PicklerHandler, ProviderDataExtractor
from nebuly.utils import is_module_version_less_than

try:
    # This import is valid only starting from openai==1.8.0
    from openai._legacy_response import (  # pylint: disable=ungrouped-imports  # type: ignore  # noqa: E501
        LegacyAPIResponse,
    )
except ImportError:

    class LegacyAPIResponse:
        pass


if is_module_version_less_than("openai", "1.21.0"):
    assistant_messages_list = "resources.beta.threads.messages.messages.Messages.list"
    assistant_messages_async_list = (
        "resources.beta.threads.messages.messages.AsyncMessages.list"
    )
else:
    assistant_messages_list = "resources.beta.threads.messages.Messages.list"
    assistant_messages_async_list = "resources.beta.threads.messages.AsyncMessages.list"


try:
    # This import is valid only starting from openai==1.8.0
    from openai._response import AsyncAPIResponse  # type: ignore  # noqa: E501
except ImportError:

    class AsyncAPIResponse:
        pass


try:
    # This import is valid only before openai==1.14.0
    from openai.types.beta.threads import ThreadMessage as Message
except ImportError:
    from openai.types.beta.threads import Message


logger = logging.getLogger(__name__)


class OpenAIPicklerHandler(PicklerHandler):
    @property
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        openai_mapping = self._object_key_attribute_mapping_openai
        azure_mapping = self._object_key_attribute_mapping_azure
        return {**openai_mapping, **azure_mapping}

    @property
    def _object_key_attribute_mapping_openai(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        names = ["api_key", "organization"]
        return {
            key: {
                "key_names": names,
                "attribute_names": names,
            }
            for key in [OpenAI, AsyncOpenAI, _ModuleClient]
        }

    @property
    def _object_key_attribute_mapping_azure(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        attribute_names = ["api_key", "organization", "base_url", "_api_version"]
        key_names = ["api_key", "organization", "base_url", "api_version"]
        return {
            key: {
                "key_names": key_names,
                "attribute_names": attribute_names,
            }
            for key in [AzureOpenAI, AsyncAzureOpenAI]
        }


def handle_openai_unpickable_objects() -> None:
    pickler_handler = OpenAIPicklerHandler()
    for obj in [AzureOpenAI, AsyncAzureOpenAI, OpenAI, AsyncOpenAI, _ModuleClient]:
        pickler_handler.handle_unpickable_object(
            obj=obj,
        )


def is_openai_generator(obj: Any) -> bool:
    return isinstance(
        obj, (openai.Stream, openai.AsyncStream)  # pylint: disable=no-member
    )


class OpenAIDataExtractor(ProviderDataExtractor):
    def __init__(
        self,
        function_name: str,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
    ):
        self.function_name = function_name
        self.original_args = original_args
        self.original_kwargs = original_kwargs

    def _extract_history(self) -> list[HistoryEntry]:
        history = self.original_kwargs.get("messages", [])[:-1]

        # Remove messages that are not from the user or the assistant
        history = [
            message
            for message in history
            if len(history) > 1 and message["role"].lower() in ["user", "assistant"]
        ]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(
                user=self._extract_content(history[i]["content"]),
                assistant=self._extract_content(history[i + 1]["content"]),
            )
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]
        return history

    @staticmethod
    def _extract_content(content: str | list[dict[str, str | dict[str, str]]]) -> str:
        if isinstance(content, str):
            # Standard case
            return content
        if isinstance(content, list):
            # Vision models case
            new_content = ""
            for item in content:
                if item.get("type") == "text":
                    if len(new_content) > 0:
                        # Add a newline between each text item
                        new_content += "\n"
                    new_content += item["text"]
            return new_content
        return json.dumps(content)

    def extract_input_and_history(self, outputs: Any) -> ModelInput:
        if self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return ModelInput(prompt=self.original_kwargs.get("prompt", ""))
        if self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            prompt = self._extract_content(
                self.original_kwargs.get("messages", [])[-1]["content"]
            )
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)
        if self.function_name in [
            assistant_messages_list,
            assistant_messages_async_list,
        ]:
            if outputs.has_more:
                return ModelInput(prompt="")
            user_messages, assistant_messages = self._openai_assistant_messages(outputs)
            input_ = user_messages[0].content[0].text.value
            return ModelInput(
                prompt=input_,
                history=[
                    HistoryEntry(input_, output)
                    for input_, output in zip(
                        [
                            message.content[0].text.value
                            for message in user_messages[1:]
                        ],
                        [
                            message.content[0].text.value
                            for message in assistant_messages[1:]
                        ],
                    )
                ],
            )

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_media(self) -> list[str] | None:
        if self.function_name in [  # pylint: disable=too-many-nested-blocks
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            content = self.original_kwargs.get("messages", [])[-1]["content"]
            media_list = []
            if isinstance(content, list):
                for item in content:
                    if item.get("type") == "image_url":
                        if isinstance(item["image_url"], str):
                            media_list.append(item["image_url"])
                        else:
                            if "url" in item["image_url"]:
                                media_list.append(item["image_url"]["url"])
                            else:
                                media_list.append(item["image_url"])
                return media_list
        return None

    def extract_output(  # pylint: disable=too-many-return-statements
        self,
        stream: bool,
        outputs: ChatCompletion
        | Completion
        | list[ChatCompletionChunk]
        | list[Completion],
    ) -> str:
        if isinstance(outputs, list) and stream:
            return self._extract_output_generator(outputs)

        if isinstance(outputs, Completion) and self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return cast(CompletionChoice, outputs.choices[0]).text
        if isinstance(outputs, ChatCompletion) and self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            if cast(Choice, outputs.choices[0]).message.content is not None:
                # Normal chat completion
                return cast(str, cast(Choice, outputs.choices[0]).message.content)
            # Chat completion with function call
            return json.dumps(
                {
                    "function_name": cast(
                        Choice, outputs.choices[0]
                    ).message.function_call.name,
                    "arguments": cast(
                        Choice, outputs.choices[0]
                    ).message.function_call.arguments,
                }
            )
        if isinstance(outputs, (APIResponse, AsyncAPIResponse, LegacyAPIResponse)):
            payload_dict = json.loads(outputs.content.decode("utf-8"))
            if payload_dict.get("object") == "chat.completion":
                payload = ChatCompletion(**payload_dict)
                return cast(str, cast(Choice, payload.choices[0]).message.content)
        if self.function_name in [
            assistant_messages_list,
            assistant_messages_async_list,
        ]:
            if outputs.has_more:
                return ""
            _, assistant_messages = self._openai_assistant_messages(outputs)

            output = assistant_messages[0].content[0].text.value
            return output

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"output type: {type(outputs)}"
        )

    def _extract_output_generator(
        self, outputs: list[Completion] | list[ChatCompletionChunk]
    ) -> str:
        if all(
            isinstance(output, Completion) for output in outputs
        ) and self.function_name in [
            "resources.completions.Completions.create",
            "resources.completions.AsyncCompletions.create",
        ]:
            return "".join(
                [getattr(output.choices[0], "text", "") or "" for output in outputs]
            )
        if all(
            isinstance(output, ChatCompletionChunk) for output in outputs
        ) and self.function_name in [
            "resources.chat.completions.Completions.create",
            "resources.chat.completions.AsyncCompletions.create",
        ]:
            # Filter only outputs where choices has at least one element
            # This is needed when using Azure OpenAI, the first chunk has no choices
            valid_outputs = [output for output in outputs if len(output.choices) > 0]
            if not all(
                getattr(output.choices[0].delta, "content") is None
                for output in valid_outputs
            ):
                # Normal chat completion
                return "".join(
                    [
                        getattr(output.choices[0].delta, "content", "") or ""
                        for output in valid_outputs
                    ]
                )
            # Chat completion with function call
            return json.dumps(
                {
                    "function_name": "".join(
                        getattr(output.choices[0].delta.function_call, "name", "") or ""
                        for output in valid_outputs
                    ),
                    "arguments": "".join(
                        getattr(output.choices[0].delta.function_call, "arguments", "")
                        or ""
                        for output in valid_outputs
                    ),
                }
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )

    def _openai_assistant_messages_join(self, messages: list[Message]) -> List[str]:
        """
        Join messages from the openai assistant when same role has multiple
        messages consecutively
        """
        joined_messages = []
        current_message = messages[0]
        for message in messages[1:]:
            if message.role == current_message.role:
                current_message.content[0].text.value = (
                    message.content[0].text.value
                    + "\n"
                    + current_message.content[0].text.value
                )
            else:
                joined_messages.append(current_message)
                current_message = message
        joined_messages.append(current_message)
        return joined_messages

    def _openai_assistant_messages(
        self, outputs: SyncCursorPage[Message]
    ) -> Tuple[List[str], List[str]]:
        """
        Extract the messages from the openai assistant

        The messages are returned in the order they were created
        """
        # We need to deepcopy the data because the data is modified in place
        messages = deepcopy(outputs.data)
        if "order" in self.original_kwargs and self.original_kwargs["order"] == "asc":
            messages = messages[::-1]

        joined_messages = self._openai_assistant_messages_join(messages)

        user_messages = [data for data in joined_messages if data.role == "user"]
        assistant_messages = [
            data for data in joined_messages if data.role == "assistant"
        ]

        return user_messages, assistant_messages



# pylint: disable=duplicate-code
import asyncio
import json
import os
from abc import ABC
from typing import Any, Optional, Sequence

from nebuly.ab_testing.types import Request, Response
from nebuly.api_key import get_api_key
from nebuly.requests import post_json_data


class ABTestingBase(ABC):
    """
    Base class for ABTesting and AsyncABTesting
    """

    def __init__(self, api_key: Optional[str] = None) -> None:
        if api_key is None:
            api_key = get_api_key()
        self._api_key = api_key
        url = os.environ.get(
            "NEBULY_AB_TESTING_API_URL",
            "https://backend.nebuly.com/api/external/variants",
        )
        self._URL = url

    @staticmethod
    def parse_raw_response(raw_response: Any) -> Response:
        """
        Parse the raw response into a Response object
        """
        json_response = json.loads(raw_response)
        reponse = Response.from_dict(json_response)
        return reponse

    @staticmethod
    def get_request_payload(user: str, feature_flags: Sequence[str]) -> str:
        request_data: Request = {
            "user": user,
            "feature_flags": feature_flags,
        }
        payload = json.dumps(request_data)
        return payload


class ABTesting(ABTestingBase):
    """
    Synchronous ABTesting class
    """

    def get_variants(self, user: str, feature_flags: Sequence[str]) -> Response:
        """
        Get the variant for each feature flag for a given user
        """
        payload = self.get_request_payload(user, feature_flags)
        raw_response = post_json_data(self._URL, payload, self._api_key)
        return self.parse_raw_response(raw_response)


class AsyncABTesting(ABTestingBase):
    """
    Asynchronous ABTesting class
    """

    async def get_variants(self, user: str, feature_flags: Sequence[str]) -> Response:
        """
        Get the variant for each feature flag for a given user
        """
        payload = self.get_request_payload(user, feature_flags)
        loop = asyncio.get_running_loop()
        raw_response = await loop.run_in_executor(
            None,
            post_json_data,
            self._URL,
            payload,
            self._api_key,
        )
        return self.parse_raw_response(raw_response)



from __future__ import annotations

import logging
import re

from nebuly.entities import HistoryEntry

logger = logging.getLogger(__name__)


def extract_anthropic_input_and_history(prompt: str) -> tuple[str, list[HistoryEntry]]:
    # Ensure that the prompt is a string following pattern "\n\nHuman:...Assistant:
    prompt = prompt.strip()
    if re.match(r"\n*Human:.*Assistant:$", prompt, re.DOTALL) is None:
        return prompt, []
    try:
        # Extract human and assistant interactions using regular expression
        pattern = re.compile(r"Human:(.*?)\n*Assistant:(.*?)(?=\n*Human:|$)", re.DOTALL)
        interactions = pattern.findall(prompt)

        # Extracting the last user input
        last_user_input = interactions[-1][0].strip()

        # Create a list of tuples for the history
        history = [
            HistoryEntry(human.strip(), assistant.strip())
            for human, assistant in interactions[:-1]
        ]

        return last_user_input, history
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("Failed to extract input and history for anthropic: %s", e)
        return prompt, []



# pylint: disable=duplicate-code, import-error, no-name-in-module
# mypy: ignore-errors
from __future__ import annotations

import json
import logging
from typing import Any

from openai.openai_object import OpenAIObject  # type: ignore

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import ProviderDataExtractor

logger = logging.getLogger(__name__)


class OpenAILegacyDataExtractor(ProviderDataExtractor):
    def __init__(
        self,
        function_name: str,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
    ):
        self.function_name = function_name
        self.original_args = original_args
        self.original_kwargs = original_kwargs

    def _extract_history(self) -> list[HistoryEntry]:
        history = self.original_kwargs.get("messages", [])[:-1]

        # Remove messages that are not from the user or the assistant
        history = [
            message
            for message in history
            if len(history) > 1 and message["role"].lower() in ["user", "assistant"]
        ]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(
                user=history[i]["content"], assistant=history[i + 1]["content"]
            )
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]
        return history

    def extract_input_and_history(self, outputs: Any) -> ModelInput:
        if self.function_name in ["Completion.create", "Completion.acreate"]:
            return ModelInput(prompt=self.original_kwargs.get("prompt", ""))
        if self.function_name in ["ChatCompletion.create", "ChatCompletion.acreate"]:
            prompt = self.original_kwargs.get("messages", [])[-1]["content"]
            history = self._extract_history()
            return ModelInput(prompt=prompt, history=history)

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: dict[str, Any]
        | list[dict[str, Any]]
        | OpenAIObject
        | list[OpenAIObject],
    ) -> str:
        if isinstance(outputs, list) and stream:
            return self._extract_output_generator(outputs)
        if isinstance(outputs, (OpenAIObject, dict)) and self.function_name in [
            "Completion.create",
            "Completion.acreate",
        ]:
            return outputs["choices"][0]["text"]  # type: ignore
        if isinstance(outputs, (OpenAIObject, dict)) and self.function_name in [
            "ChatCompletion.create",
            "ChatCompletion.acreate",
        ]:
            if outputs["choices"][0]["message"].get("content") is not None:
                # Normal chat completion
                return outputs["choices"][0]["message"]["content"]  # type: ignore
            # Function call
            return json.dumps(
                {
                    "function_name": outputs["choices"][0]["message"]["function_call"][
                        "name"
                    ],
                    "arguments": outputs["choices"][0]["message"]["function_call"][
                        "arguments"
                    ],
                }
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"type of outputs: {type(outputs)}"
        )

    def _extract_output_generator(
        self, outputs: list[OpenAIObject] | list[dict[str, Any]]
    ) -> str:
        if all(
            isinstance(output, (OpenAIObject, dict)) for output in outputs
        ) and self.function_name in ["Completion.create", "Completion.acreate"]:
            return "".join([output["choices"][0].get("text", "") for output in outputs])
        if all(
            isinstance(output, (OpenAIObject, dict)) for output in outputs
        ) and self.function_name in ["ChatCompletion.create", "ChatCompletion.acreate"]:
            if not all(
                output["choices"][0]["delta"].get("content") is None
                for output in outputs
            ):
                # Normal chat completion
                return "".join(
                    [
                        output["choices"][0]["delta"].get("content", "")
                        for output in outputs
                    ]
                )
            # Function call
            return json.dumps(
                {
                    "function_name": "".join(
                        output["choices"][0]["delta"]
                        .get("function_call", {})
                        .get("name", "")
                        for output in outputs
                    ),
                    "arguments": "".join(
                        output["choices"][0]["delta"]
                        .get("function_call", {})
                        .get("arguments", "")
                        for output in outputs
                    ),
                }
            )

        raise ValueError(
            f"Unknown function name: {self.function_name} or "
            f"type of outputs: {type(outputs)}"
        )



from __future__ import annotations

import logging
import uuid
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Sequence, Tuple, cast
from uuid import UUID

import langchain
from langchain.callbacks.base import BaseCallbackHandler
from langchain.chains.base import Chain
from langchain.load import load
from langchain.prompts import ChatPromptTemplate, PromptTemplate
from langchain.prompts.chat import AIMessagePromptTemplate, HumanMessagePromptTemplate
from langchain.schema import Document
from langchain.schema.messages import AIMessage, BaseMessage, HumanMessage, ToolMessage
from langchain.schema.output import LLMResult
from langchain.schema.runnable import RunnableSequence

from nebuly.entities import HistoryEntry, InteractionWatch, ModelInput, SpanWatch
from nebuly.events import Event, EventData, EventsStorage
from nebuly.requests import post_message

logger = logging.getLogger(__name__)


def _get_spans(events: dict[UUID, Event]) -> list[SpanWatch]:
    return [event.as_span_watch() for event in events.values()]


def _process_prompt_template(
    inputs: dict[str, Any] | Any, prompt: PromptTemplate
) -> str:
    if isinstance(inputs, dict):
        return prompt.format(**{key: inputs.get(key) for key in prompt.input_variables})
    return prompt.format(**{prompt.input_variables[0]: inputs})


def _process_chat_prompt_template(
    inputs: dict[str, Any] | Any, prompt: ChatPromptTemplate
) -> tuple[str, list[HistoryEntry]]:
    messages = []
    for message in prompt.messages:
        if isinstance(inputs, dict):
            input_vars = {key: inputs.get(key) for key in message.input_variables}  # type: ignore  # noqa: E501  # pylint: disable=line-too-long
        else:
            input_vars = {message.input_variables[0]: inputs}  # type: ignore
        if isinstance(message, (HumanMessagePromptTemplate, AIMessagePromptTemplate)):
            messages.append((message.format(**input_vars).content))
    last_prompt = str(messages[-1])
    message_history = messages[:-1]

    if len(message_history) % 2 != 0:
        logger.warning("Odd number of chat history elements, ignoring last element")
        message_history = message_history[:-1]

    # Convert the history to [(user, assistant), ...] format
    history = [
        HistoryEntry(
            user=str(message_history[i]), assistant=str(message_history[i + 1])
        )
        for i in range(0, len(message_history), 2)
        if i < len(message_history) - 1
    ]

    return last_prompt, history


def _get_input_and_history(chain: Chain, inputs: dict[str, Any] | Any) -> ModelInput:
    chains = getattr(chain, "chains", None)
    if chains is not None:
        # If the chain is a SequentialChain, we need to get the
        # prompt from the first chain
        prompt = getattr(chains[0], "prompt", None)
    else:
        # If the chain is not a SequentialChain, we need to get
        # the prompt from the chain
        prompt = getattr(chain, "prompt", None)
        if prompt is None:
            if not isinstance(inputs, dict):
                return ModelInput(prompt=inputs)
            return ModelInput(prompt=inputs["input"])

    if isinstance(prompt, PromptTemplate):
        return ModelInput(prompt=_process_prompt_template(inputs, prompt))

    if isinstance(prompt, ChatPromptTemplate):
        prompt, history = _process_chat_prompt_template(inputs, prompt)
        return ModelInput(prompt=prompt, history=history)

    raise ValueError(f"Unknown prompt type: {prompt}")


def _get_input_and_history_runnable_seq(
    sequence: RunnableSequence[Any, Any], inputs: dict[str, Any] | Any
) -> ModelInput:
    steps = getattr(sequence, "steps", None)

    if steps is not None and len(steps) > 0:
        for step in steps:
            if isinstance(step, PromptTemplate):
                return ModelInput(prompt=_process_prompt_template(inputs, step))

            if isinstance(step, ChatPromptTemplate):
                prompt, history = _process_chat_prompt_template(inputs, step)
                return ModelInput(prompt=prompt, history=history)

    return ModelInput(prompt="")


def _get_output_chain(chain: Chain, result: dict[str, Any]) -> str:
    if len(chain.output_keys) == 1:
        return str(result[chain.output_keys[0]])
    output = {}
    for key in chain.output_keys:
        output[key] = result[key]
    return _parse_output(output)


def _parse_output(output: str | dict[str, Any] | AIMessage) -> str:
    if isinstance(output, dict):
        if "answer" in output:
            return str(output["answer"])
        if "output" in output:
            return str(output["output"])
        return "\n".join([f"{key}: {value}" for key, value in output.items()])
    if isinstance(output, AIMessage):
        return str(output.content)
    return output


def _parse_langchain_data(  # pylint: disable=too-many-return-statements
    data: Any,
) -> str:
    if isinstance(data, dict):
        if len(data) == 1 and isinstance(list(data.values())[0], str):
            return str(list(data.values())[0])
        if "answer" in data:
            # If the data is a retrieval chain, we want to return the answer
            return str(data["answer"])
        if "output" in data:
            # If the data is a retrieval chain, we want to return the output
            return str(data["output"])
        if "input" in data:
            # Needed when calling invoke with input and chat_history
            return str(data["input"])
        if "question" in data:
            # Needed when calling invoke with question and chat_history
            return str(data["question"])
        if "messages" in data and isinstance(
            data["messages"][-1], (HumanMessage, AIMessage, ToolMessage)
        ):
            return str(data["messages"][-1].content)
        return "\n".join([f"{key}: {value}" for key, value in data.items()])
    if isinstance(data, list):
        data = data[-1]
    if isinstance(data, (HumanMessage, AIMessage, ToolMessage)):
        return str(data.content)
    return str(data)


def _parse_langchain_history(inputs: dict[str, Any]) -> list[HistoryEntry]:
    if "chat_history" in inputs:
        history = inputs["chat_history"]
        if len(history) > 0:
            if isinstance(history[0], tuple):
                # In some chains, the chat history is a list of tuples
                return [
                    HistoryEntry(user=str(message[0]), assistant=str(message[1]))
                    for message in history
                ]

            human_messages = [message for message in history if message.type == "human"]
            ai_messages = [message for message in history if message.type == "ai"]

            if len(human_messages) != len(ai_messages):
                logger.warning(
                    "Unequal number of human and AI messages in chat history"
                )

            return [
                HistoryEntry(user=str(human.content), assistant=str(ai.content))
                for human, ai in zip(human_messages, ai_messages)
            ]
    return []


class EventType(Enum):
    """
    The type of event generated by LangChain.
    """

    CHAIN = "chain"
    TOOL = "tool"
    RETRIEVAL = "retrieval"
    LLM_MODEL = "llm_model"
    CHAT_MODEL = "chat_model"


@dataclass
class LangChainEvent(Event):
    @property
    def input(self) -> str:
        if self.data.kwargs is None:
            raise ValueError("Event has no kwargs.")
        if self.data.type is EventType.CHAT_MODEL:
            messages = self.data.kwargs.get("messages")
            if messages is None:
                raise ValueError("Event has no messages.")
            return str(messages[-1][-1].content)
        if self.data.type is EventType.LLM_MODEL:
            prompts = self.data.kwargs.get("prompts")
            if prompts is None:
                raise ValueError("Event has no prompts.")
            return str(prompts[-1])
        if self.data.type is EventType.CHAIN:
            inputs = self.data.kwargs.get("inputs")
            if inputs is None:
                raise ValueError("Event has no inputs.")
            try:
                chain = load(self.data.kwargs.get("serialized", {}))
                if isinstance(chain, RunnableSequence):
                    return _get_input_and_history_runnable_seq(chain, inputs).prompt
                return _get_input_and_history(chain, inputs).prompt
            except NotImplementedError:
                return _parse_langchain_data(inputs)

        raise ValueError(f"Event type {self.data.type} not supported.")

    @property
    def output(self) -> str:
        if self.data.output is None:
            raise ValueError("Event has no output.")
        if self.data.type in [EventType.CHAT_MODEL, EventType.LLM_MODEL]:
            return str(self.data.output.generations[-1][0].text)
        if self.data.type is EventType.CHAIN:
            if self.data.kwargs is None:
                raise ValueError("Event has no kwargs.")
            try:
                chain = load(self.data.kwargs.get("serialized", {}))
                if isinstance(chain, RunnableSequence):
                    return _parse_output(self.data.output)
                return _get_output_chain(chain, self.data.output)
            except NotImplementedError:
                return _parse_langchain_data(self.data.output)

        raise ValueError(f"Event type {self.data.type} not supported.")

    @property
    def history(self) -> list[HistoryEntry]:
        if self.data.kwargs is None:
            raise ValueError("Event has no kwargs.")

        if self.data.type is EventType.CHAT_MODEL:
            messages = self.data.kwargs.get("messages")
            if messages is None:
                raise ValueError("Event has no messages.")
            history = [
                message
                for message in messages[-1][:-1]
                if isinstance(message, (HumanMessage, AIMessage))
            ]
            if len(history) % 2 != 0:
                raise ValueError(
                    "Odd number of chat history elements, please provide "
                    "a valid history."
                )
            if len(history) == 0:
                return []
            return [
                HistoryEntry(
                    user=str(history[i].content), assistant=str(history[i + 1].content)
                )
                for i in range(0, len(history) - 1, 2)
            ]
        if self.data.type is EventType.LLM_MODEL:
            return []
        if self.data.type is EventType.CHAIN:
            inputs = self.data.kwargs.get("inputs")
            if inputs is None:
                raise ValueError("Event has no inputs.")
            try:
                chain = load(self.data.kwargs.get("serialized", {}))
                if isinstance(chain, RunnableSequence):
                    return _get_input_and_history_runnable_seq(chain, inputs).history
                return _get_input_and_history(chain, inputs).history
            except NotImplementedError:
                return _parse_langchain_history(inputs)

        raise ValueError(f"Event type {self.data.type} not supported.")

    def as_span_watch(self) -> SpanWatch:
        if self.end_time is None:
            raise ValueError("Event has not been finished yet.")
        return SpanWatch(
            span_id=self.event_id,
            module=self.module,
            version="unknown",
            function=self._get_function(),
            called_start=self.start_time,
            called_end=self.end_time,
            called_with_args=cast(Tuple[Any], self.data.args),
            called_with_kwargs=cast(Dict[str, Any], self.data.kwargs),
            returned=self.data.output,
            generator=False,
            generator_first_element_timestamp=None,
            provider_extras={
                "event_type": self.data.type.value,
            },
            rag_source=self._get_rag_source(),
        )

    def _get_function(self) -> str:
        if self.data.kwargs is None or len(self.data.kwargs) == 0:
            raise ValueError("Event has no kwargs.")
        if self.data.type is EventType.TOOL:
            return self.data.kwargs["serialized"]["name"]  # type: ignore
        return ".".join(self.data.kwargs["serialized"]["id"])

    def _get_rag_source(self) -> str | None:
        if self.data.kwargs is None or len(self.data.kwargs) == 0:
            raise ValueError("Event has no kwargs.")
        if self.data.type is EventType.TOOL:
            return self.data.kwargs["serialized"]["name"]  # type: ignore
        if self.data.type is EventType.RETRIEVAL:
            return self.data.kwargs["serialized"]["id"][-1]  # type: ignore

        return None


class LangChainTrackingHandler(BaseCallbackHandler):  # noqa
    def __init__(  # pylint: disable=too-many-arguments
        self,
        api_key: str,
        user_id: str,
        user_group_profile: str | None = None,
        feature_flags: list[str] | None = None,
        nebuly_tags: dict[str, str] | None = None,
    ) -> None:
        self._check_version()
        self.api_key = api_key
        self.nebuly_user = user_id
        self.nebuly_user_group = user_group_profile
        self.nebuly_feature_flags = feature_flags
        self.nebuly_tags = nebuly_tags
        self._events_storage = EventsStorage()

    @staticmethod
    def _check_version() -> None:
        if langchain.__version__ < "0.1.0":
            logger.warning(
                "LangChain version 0.1.0 or higher is required to use the "
                "LangChainTrackingHandler with Nebuly SDK. "
                "Please upgrade your LangChain version.",
            )

    def _send_interaction(self, run_id: uuid.UUID) -> None:
        if (
            len(self._events_storage.events) < 1
            or run_id not in self._events_storage.events
        ):
            raise ValueError(f"Event {run_id} not found in events storage.")

        interaction = InteractionWatch(
            end_user=self.nebuly_user,
            end_user_group_profile=self.nebuly_user_group,
            input=self._events_storage.events[run_id].input,
            output=self._events_storage.events[run_id].output,
            time_start=self._events_storage.events[run_id].start_time,
            time_end=cast(datetime, self._events_storage.events[run_id].end_time),
            history=self._events_storage.events[run_id].history,
            spans=_get_spans(events=self._events_storage.events),
            hierarchy={
                event.event_id: event.hierarchy.parent_run_id
                if event.hierarchy is not None
                else None
                for event in self._events_storage.events.values()
            },
            feature_flags=self.nebuly_feature_flags,
            tags=self.nebuly_tags,
        )
        post_message(interaction, self.api_key)

    def on_tool_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        input_str: str,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.TOOL,
            kwargs={
                "serialized": serialized,
                "input_str": input_str,
                **kwargs,
            },
        )
        self._events_storage.add_event(
            LangChainEvent, run_id, parent_run_id, data, module="langchain"
        )

    def on_tool_end(  # pylint: disable=arguments-differ
        self,
        output: str,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self._events_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=output
        )
        self._events_storage.events[run_id].set_end_time()

    def on_retriever_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        query: str,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.RETRIEVAL,
            kwargs={
                "serialized": serialized,
                "query": query,
                **kwargs,
            },
        )
        self._events_storage.add_event(
            LangChainEvent, run_id, parent_run_id, data, module="langchain"
        )

    def on_retriever_end(  # pylint: disable=arguments-differ
        self,
        documents: Sequence[Document],
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self._events_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=documents
        )
        self._events_storage.events[run_id].set_end_time()

    def on_llm_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.LLM_MODEL,
            kwargs={
                "serialized": serialized,
                "prompts": prompts,
                **kwargs,
            },
        )
        self._events_storage.add_event(
            LangChainEvent, run_id, parent_run_id, data, module="langchain"
        )

    def on_llm_end(  # pylint: disable=arguments-differ
        self,
        response: LLMResult,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self._events_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=response
        )
        self._events_storage.events[run_id].set_end_time()

        if len(self._events_storage.events) == 1:
            self._send_interaction(run_id)
            self._events_storage.delete_events(run_id)

    def on_chat_model_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.CHAT_MODEL,
            kwargs={
                "serialized": serialized,
                "messages": messages,
                **kwargs,
            },
        )
        self._events_storage.add_event(
            LangChainEvent, run_id, parent_run_id, data, module="langchain"
        )

    def on_chain_start(  # pylint: disable=arguments-differ
        self,
        serialized: dict[str, Any],
        inputs: dict[str, Any],
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        data = EventData(
            type=EventType.CHAIN,
            kwargs={
                "serialized": serialized,
                "inputs": inputs,
                **kwargs,
            },
        )
        self._events_storage.add_event(
            LangChainEvent, run_id, parent_run_id, data, module="langchain"
        )

    def on_chain_end(  # pylint: disable=arguments-differ
        self,
        outputs: dict[str, Any],
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        self._events_storage.events[run_id].data.add_end_event_data(
            kwargs=kwargs, output=outputs
        )
        self._events_storage.events[run_id].set_end_time()

        if self._events_storage.events[run_id].hierarchy is None:
            self._send_interaction(run_id)
            self._events_storage.delete_events(run_id)



from __future__ import annotations

import json
import logging
import os
import urllib.request
from time import sleep
from typing import Any

from nebuly.entities import InteractionWatch
from nebuly.exceptions import InvalidNebulyKeyError

logger = logging.getLogger(__name__)


class CustomJSONEncoder(json.JSONEncoder):
    ACCEPTED_KEYS = ["_model_id", "_model", "_context", "_examples", "_message_history"]

    @staticmethod
    def _check_key(key: str) -> bool:
        """
        Reject all private keys except for ones that contain info about the model
        """
        return not key.startswith("_") or key in CustomJSONEncoder.ACCEPTED_KEYS

    @staticmethod
    def _check_value(value: Any) -> bool:
        correct_types: tuple[type[Any], ...] = (
            str,
            int,
            float,
            bool,
            list,
            tuple,
            dict,
            type(None),
        )
        has_correct_type = isinstance(value, correct_types)
        is_valid_class = False
        try:
            if hasattr(value, "__dict__"):
                json.dumps(value.__dict__)
                is_valid_class = True
        except TypeError:
            is_valid_class = any(
                (hasattr(value, key) for key in CustomJSONEncoder.ACCEPTED_KEYS)
            )
        return has_correct_type or is_valid_class

    def default(self, o: Any) -> Any:
        if isinstance(o, bytes):
            return o.decode("utf-8")
        if hasattr(o, "to_dict"):
            return o.to_dict()
        try:
            return json.JSONEncoder.default(self, o)
        except Exception:  # pylint: disable=broad-except
            if hasattr(o, "__dict__"):
                dict_repr = o.__dict__
                return {
                    key: dict_repr[key]
                    for key in dict_repr
                    if self._check_key(key) and self._check_value(dict_repr[key])
                }
            return str(o)


def post_message(
    watched: InteractionWatch, api_key: str, anonymize: bool = True
) -> None:
    message = json.dumps(
        {"body": watched, "anonymize": anonymize}, cls=CustomJSONEncoder
    )
    url = os.environ.get(
        "NEBULY_API_URL",
        "https://backend.nebuly.com/event-ingestion/api/v1/events",
    )
    if watched.api_key:
        api_key = watched.api_key
    post_json_data(url, message, api_key)


def post_json_data(url: str, json_data: str, api_key: str) -> Any:
    request = urllib.request.Request(
        url,
        data=json_data.encode("utf-8"),
        headers={
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        },
    )

    # TODO: move retries out of this function
    tries = 0
    while tries < 3:
        try:
            with urllib.request.urlopen(request, timeout=3) as response:  # nosec
                response_body = response.read().decode("utf-8")
                logger.debug("response_body: %s", response_body)
                return response_body
        except urllib.error.HTTPError as e:
            if e.code == 401:
                raise InvalidNebulyKeyError(
                    "Invalid Nebuly API key, please provide a valid one and try again."
                ) from e
            logger.exception(
                "HTTPError when publishing the interaction: %s, "
                "retrying in 3 seconds...",
                e,
            )
            sleep(3)
            tries += 1
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(
                "Error when publishing the interaction: %s, "
                "retrying in 3 seconds...",
                e,
            )
            sleep(3)
            tries += 1

    logger.error("Failed to publish the interaction, giving up.")
    return None



from __future__ import annotations

import abc
import copyreg
from typing import Any, Callable

from nebuly.entities import ModelInput


class ProviderDataExtractor(abc.ABC):
    @abc.abstractmethod
    def __init__(
        self,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
        function_name: str,
    ):
        raise NotImplementedError("init not implemented")

    @abc.abstractmethod
    def extract_input_and_history(self, outputs: Any) -> ModelInput | list[ModelInput]:
        raise NotImplementedError("extract_input_and_history not implemented")

    @abc.abstractmethod
    def extract_output(self, stream: bool, outputs: Any) -> str | list[str]:
        raise NotImplementedError("extract_output not implemented")

    def extract_media(self) -> list[str] | None:
        return None


class PicklerHandler(abc.ABC):
    @property
    @abc.abstractmethod
    def _object_key_attribute_mapping(
        self,
    ) -> dict[Callable[[Any], Any], dict[str, list[str]]]:
        raise NotImplementedError("This method should be implemented by subclasses")

    def _get_attribute(self, obj: Any, attribute: str) -> Any:
        if "." not in attribute:
            return getattr(obj, attribute)

        attributes = attribute.split(".")
        return self._get_attribute(
            getattr(obj, attributes[0]), ".".join(attributes[1:])
        )

    @staticmethod
    def _unpickle_object(
        constructor: Callable[[Any], Any], kwargs: dict[str, Any]
    ) -> Any:
        return constructor(**kwargs)  # type: ignore

    def _pickle_object(
        self, obj: Any
    ) -> tuple[
        Callable[[Callable[[Any], Any], dict[str, Any]], Any],
        tuple[Callable[[Any], Any], dict[str, Any]],
    ]:
        return self._unpickle_object, (
            obj.__class__,
            {
                key: self._get_attribute(obj, attribute)
                for key, attribute in zip(
                    self._object_key_attribute_mapping[obj.__class__]["key_names"],
                    self._object_key_attribute_mapping[obj.__class__][
                        "attribute_names"
                    ],
                )
            },
        )

    def handle_unpickable_object(self, obj: Callable[[Any], Any]) -> None:
        copyreg.pickle(obj, self._pickle_object)  # type: ignore



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional


@dataclass
class FeedbackActionMetadata:
    input: Optional[str]
    output: Optional[str]
    end_user: str
    timestamp: datetime
    anonymize: bool
    end_user_group_profile: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "input": self.input,
            "output": self.output,
            "end_user": self.end_user,
            "end_user_group_profile": self.end_user_group_profile,
            "timestamp": self.timestamp.isoformat(),
            "anonymize": self.anonymize,
        }


class FeedbackActionName(str, Enum):
    THUMBS_UP = "thumbs_up"
    THUMBS_DOWN = "thumbs_down"
    COPY_INPUT = "copy_input"
    COPY_OUTPUT = "copy_output"
    PASTE = "paste"


@dataclass
class FeedbackAction:
    slug: FeedbackActionName
    extras: Optional[Dict[str, Any]] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "slug": self.slug,
            **(self.extras or {}),
        }



# pylint: disable=duplicate-code
from __future__ import annotations

import logging
from typing import Any, Dict, List, cast

from transformers.pipelines import (  # type: ignore
    Conversation,
    ConversationalPipeline,
    TextGenerationPipeline,
)
from typing_extensions import TypeAlias

from nebuly.entities import HistoryEntry, ModelInput
from nebuly.providers.base import ProviderDataExtractor

logger = logging.getLogger(__name__)


TextGeneration: TypeAlias = List[Dict[str, str]]


class HuggingFaceDataExtractor(ProviderDataExtractor):
    def __init__(
        self,
        function_name: str,
        original_args: tuple[Any, ...],
        original_kwargs: dict[str, Any],
    ):
        self.function_name = function_name
        self.original_args = original_args
        self.original_kwargs = original_kwargs

    @staticmethod
    def _extract_history(conversations: Conversation) -> list[HistoryEntry]:
        history = conversations.messages[:-1]

        # Remove messages that are not from the user or the assistant
        history = [
            message
            for message in history
            if len(history) > 1 and message["role"].lower() in ["user", "assistant"]
        ]

        if len(history) % 2 != 0:
            logger.warning("Odd number of chat history elements, ignoring last element")
            history = history[:-1]

        # Convert the history to [(user, assistant), ...] format
        history = [
            HistoryEntry(
                user=history[i]["content"], assistant=history[i + 1]["content"]
            )
            for i in range(0, len(history), 2)
            if i < len(history) - 1
        ]

        return history  # type: ignore

    def extract_input_and_history(self, outputs: Any) -> ModelInput | list[ModelInput]:
        pipeline: ConversationalPipeline | TextGenerationPipeline = self.original_args[
            0
        ]  # noqa: E501
        if isinstance(pipeline, ConversationalPipeline):
            conversations: Conversation | list[Conversation] = self.original_args[1]
            if isinstance(conversations, Conversation):
                prompt = conversations.messages[-1]["content"]
                history = self._extract_history(conversations)
                return ModelInput(prompt=prompt, history=history)
            return [
                ModelInput(
                    prompt=conversation.messages[-1]["content"],
                    history=self._extract_history(conversation),
                )
                for conversation in conversations
            ]
        if isinstance(pipeline, TextGenerationPipeline):
            prompts: str | list[str] = self.original_args[1]
            if isinstance(prompts, str):
                return ModelInput(prompt=prompts)
            return [ModelInput(prompt=prompt) for prompt in prompts]

        raise ValueError(f"Unknown function name: {self.function_name}")

    def extract_output(
        self,
        stream: bool,
        outputs: Conversation
        | list[Conversation]
        | TextGeneration
        | list[TextGeneration],
    ) -> str | list[str]:
        if stream:
            return self._extract_output_generator(outputs)

        if isinstance(outputs, Conversation):
            return cast(str, outputs.generated_responses[-1])
        if isinstance(outputs, list):
            if all(isinstance(output, Conversation) for output in outputs):
                return [out.generated_responses[-1] for out in outputs]  # type: ignore
            if all(isinstance(output, dict) for output in outputs):  # TextGeneration
                return outputs[0]["generated_text"]  # type: ignore
            if all(  # list[TextGeneration]
                isinstance(output, list) for output in outputs
            ) and all(
                isinstance(output[0], dict) for output in outputs  # type: ignore
            ):
                return [out[0]["generated_text"] for out in outputs]  # type: ignore

        raise ValueError(
            f"Unknown function name: {self.function_name} "
            f"or output type: {type(outputs)}"
        )

    def _extract_output_generator(self, outputs: Any) -> str:
        raise NotImplementedError("HuggingFace does not support streaming yet")
